import json
import base64
import logging
from typing import Optional, TypeVar, Type, Generic

T = TypeVar('T')


class TaekusCommunicationModelBase:
    pass


class TaekusCommunicationModel(Generic[T]):

    def to_encoded_message(self) -> str:
        return to_message(self)

    @classmethod
    def to_model(cls: Type[T], input: str) -> Optional[T]:
        return convert_to(input, cls)


def encode(value: str) -> str:
    return base64.b64encode(value.strip().encode()).decode()


def convert_to(input: str, model_class: Type[T]) -> Optional[T]:
    logger = logging.getLogger('communication_writer')
    try:
        decoded = base64.b64decode(input).decode()
        return model_class(**json.loads(decoded))
    except Exception as ex:
        try:
            decoded = base64.b64decode(input).decode()
            logger.error(f"Message could not be converted to model as there was a mismatch: {decoded}", exc_info=True)
        except Exception as ex:
            logger.error(f"Message sent was not base64 decoded. value={input}", exc_info=True)
    return None


def to_message(value: T) -> str:
    if not isinstance(value, TaekusCommunicationModelBase):
        raise TypeError('value must be a TaekusCommunicationModelBase')
    return encode(json.dumps(value.__dict__))



import uuid
from dataclasses import dataclass

from .communication_model import TaekusCommunicationModelBase, TaekusCommunicationModel


@dataclass
class TestPublishMessageModel(TaekusCommunicationModelBase):
    job_id: str
    message: str


class TestPublishMessage(TaekusCommunicationModel):
    queue_name: str = "TEST-PUBLISH-QUEUE"



import json
from dataclasses import dataclass, asdict


@dataclass
class CommunicationWrapper:
    common_id: str
    content: str

    def to_json(self) -> str:
        return json.dumps(asdict(self))

    @classmethod
    def from_json(cls, json_str: str) -> "CommunicationWrapper":
        data = json.loads(json_str)
        return cls(**data)



import jwt
import json
from tamqp.models.communication_wrapper import CommunicationWrapper


class EncodedProvider:
    def __init__(self, key: str) -> None:
        self.key: str = key
        self.algorithm = "HS256"

    def encode(self, message: str, common_id: str) -> str:
        payload = {"CommonId": common_id, "Content": message}
        return jwt.encode(payload, self.key, algorithm=self.algorithm)

    def decode(self, message: str) -> CommunicationWrapper:
        try:
            payload = jwt.decode(message, self.key, algorithms=[self.algorithm])
            common_id = payload["CommonId"]
            content = payload["Content"]
            return CommunicationWrapper(common_id, content)
        except jwt.ExpiredSignatureError:
            raise Exception("Invalid message: Signature has expired.")
        except jwt.InvalidTokenError:
            raise Exception("Invalid message: Token verification failed.")
        except KeyError:
            raise Exception("Invalid message: Missing required payload fields.")



import logging
from typing import List
from tamqp.models.amqp_init import AMQPInit
from tamqp.services.amqp_provider import AMQPProvider


class AMQP(AMQPProvider):
    def __init__(self, parameters: AMQPInit, subscriptions=None) -> None:
        super().__init__(parameters, key="taekus_key")
        if subscriptions is None:
            subscriptions = []
        self.parameters = parameters
        self.subscriptions = subscriptions

        self.setup_retry_infrastructure()

    def subscribe(self) -> List[None]:
        results = []
        for queue in self.subscriptions:
            while True:
                try:
                    self.setup_queue(queue)
                    logging.info(
                        f"Successfully initialized - channelId={self.consumer_channel.channel_number}  queue={queue.name}")
                    results.append(None)
                    break  # Break the loop if the operation was successful.
                except Exception as ex:
                    logging.error(
                        f"Failed to subscribe - channelId={self.consumer_channel.channel_number}  queue={queue.name} error={ex}")
                    self.establish_connection_and_channels()  # Reconnect and try again.
        return results


    def close(self) -> None:
        try:
            for _, channel in self.channels.items():
                if channel.is_open:
                    channel.close()

            for _, conn in self.connections.items():
                if conn.is_open:
                    conn.close(self.CLOSING_TIMEOUT)

            logging.info("[AMQP] Closed Connection - Successfully")
        except Exception as ex:
            logging.error(f"[AMQP] Error Closing Connection - error={ex}")




import logging
from typing import Callable
from pika.spec import Basic

from .models.communication_model import convert_to
from .models.test_model import TestMessageModel
from ..models.communication_wrapper import CommunicationWrapper
from ..services.amqp import AMQP


class JobConsumer:

    def __init__(self):
        self.logger = logging.getLogger('job_consumer')

    def handle_scrape_flight(self, rmq: AMQP) -> Callable[[CommunicationWrapper, Basic.Deliver], None]:
        def callback(wrapper: CommunicationWrapper, tag: Basic.Deliver) -> None:
            model = convert_to(wrapper.content, TestMessageModel)

            if model is not None:
                message = model.message
                if message:
                    self.logger.info(f"Received message to scrape flight for message={message}")
                    try:
                        # TODO: Create orchestration layer here to kick off scraper
                        print(f"Triggered flight scrape job for message={message}")
                        self.logger.info(f"Triggered flight scrape job for message={message}")
                        rmq.ack_message(tag)
                    except (ValueError, Exception) as e:
                        self.logger.error(f"Error in triggering flight scrape job for message={message}", e)
                        rmq.ack_message(tag)
                    return
                else:
                    self.logger.error("Message field is empty in TestMessageModel instance.")
            else:
                self.logger.error(
                    "Could not convert the dequeued message into a model. This is most likely a malformed "
                    "message.")
            rmq.ack_message(tag)

        return callback



from typing import Optional


class AMQPInit:
    def __init__(
            self,
            username: str,
            password: str,
            hostname: str,
            port: int = 5672,
            virtual_host: Optional[str] = "/",
            encryption_key: Optional[str] = "taekus_key",
            is_tls: bool = False
    ) -> None:
        if not username:
            raise ValueError("username is null or empty")
        if not password:
            raise ValueError("password is null or empty")
        if not hostname:
            raise ValueError("hostname is null or empty")

        self.username = username
        self.password = password
        self.hostname = hostname
        self.port = port
        self.virtual_host = virtual_host
        self.encryption_key = encryption_key
        self.is_tls = is_tls

    @classmethod
    def from_args(cls, username: str, password: str, hostname: str, port: int, virtual_host: str,
                  encryption_key: str) -> "AMQPInit":
        vh = virtual_host if virtual_host.startswith("/") else f"/{virtual_host}"
        return cls(username, password, hostname, port, virtual_host=vh, encryption_key=encryption_key)

    @classmethod
    def from_args_and_virtual_host(cls, username: str, password: str, hostname: str, port: int,
                                   virtual_host: str) -> "AMQPInit":
        vh = virtual_host if virtual_host.startswith("/") else f"/{virtual_host}"
        return cls(username, password, hostname, port, virtual_host=vh)



from typing import List

from tamqp.models.retry_queue import RetryQueue


class AMQPProviderConfig:
    RETRY_COUNT_HEADER = "x-retries"
    X_MATCH = "x-match"
    X_MESSAGE_TTL = "x-message-ttl"
    X_DEAD_LETTER_EXCHANGE = "x-dead-letter-exchange"
    DEFAULT_EXCHANGE = ""
    DEFAULT_ROUTING_KEY = ""
    DEFAULT_TAEKUS_KEY = "taekus_key"
    RETRY_EXCHANGE = "retry.exchange"
    DEAD_LETTER_QUEUE = "retry.dead_letter"
    SENSITIVE_CONTENT_KEY = "SensitiveContent"
    SENSITIVE_REDACTED_MESSAGE = "Message was redacted due to sensitive information."

    MAX_NO_OF_RETRIES: int = 25
    CLOSING_TIMEOUT: int = int(30000)
    RETRY_QUEUES: List[RetryQueue] = [
        RetryQueue(1, 10000),
        RetryQueue(2, 60000),
        RetryQueue(3, 300000),
        RetryQueue(4, 900000),
        RetryQueue(5, 1800000)
    ]



import base64


class StringDecoder:
    class _StringDecoder:
        def __init__(self, value: str) -> None:
            self._value: str = value

        def decode(self) -> str:
            try:
                decoded_bytes = base64.b64decode(self._value)
                decoded_str = decoded_bytes.decode("utf-8")
                return decoded_str
            except Exception:
                return self._value

    def __init__(self, value: str) -> None:
        self._value: str = value

    def decode(self) -> str:
        return self._StringDecoder(self._value).decode()



import uuid
from dataclasses import dataclass

from .communication_model import TaekusCommunicationModelBase, TaekusCommunicationModel


@dataclass
class TestMessageModel(TaekusCommunicationModelBase):
    job_id: str
    message: str


class TestMessage(TaekusCommunicationModel):
    queue_name: str = "TEST-QUEUE"



import time
import pika
import uuid
from typing import Dict, Optional

from tamqp.models.amqp_init import AMQPInit
from tamqp.models.rabbit_library import Exchange, QueueSubscribed, ExchangeType
from tamqp.models.retry_queue import RetryQueue
from tamqp.services.amqp_provider_config import AMQPProviderConfig
from tamqp.services.amqp_utils import AMQPUtils
from tamqp.services.encoded_provider import EncodedProvider
from tamqp.utils.string_decoder import StringDecoder
from pika.spec import BasicProperties
from pika.adapters.blocking_connection import BlockingChannel


class AMQPProvider(EncodedProvider, AMQPProviderConfig, AMQPUtils):
    def __init__(self, parameters: AMQPInit, key: str) -> None:
        super().__init__(key)
        self.content_type_text = "text/plain"
        self.parameters: AMQPInit = parameters
        self.destination_declared: bool = False
        self.key: str = parameters.encryption_key

        self.connections: Dict[str, Optional[pika.BlockingConnection]] = {"consumer": None, "publisher": None}
        self.channels: Dict[str, Optional[BlockingChannel]] = {"consumer": None, "publisher": None}

        self.consumer_channel = self.channels["consumer"]
        self.publisher_channel = self.channels["publisher"]

        self.establish_connection_and_channels()

    def establish_connection_and_channels(self):
        while True:
            try:
                self.connections["consumer"] = self.init_connection(self.parameters)
                self.connections["publisher"] = self.init_connection(self.parameters)
                self.channels["consumer"] = self.connections["consumer"].channel()
                self.channels["publisher"] = self.connections["publisher"].channel()
                self.consumer_channel = self.channels["consumer"]
                self.publisher_channel = self.channels["publisher"]
                break
            except Exception as e:
                print(f"Error while establishing connection: {e}, retrying...")
                time.sleep(5)  # Wait for 5 seconds before retrying

    def setup_retry_queue(self, queue: RetryQueue) -> None:
        queue_props = self.generate_props({AMQPProviderConfig.X_MESSAGE_TTL: queue.ttl, AMQPProviderConfig.X_DEAD_LETTER_EXCHANGE: queue.dlx_exchange})
        self.consumer_channel.queue_declare(queue.return_name, True, False, False, False, queue_props)

    def setup_retry_infrastructure(self) -> None:
        self.consumer_channel.exchange_declare(AMQPProviderConfig.RETRY_EXCHANGE, ExchangeType.DIRECT.value, True, False,
                                               arguments=self.generate_props({}))
        for queue in self.RETRY_QUEUES:
            self.setup_retry_queue(queue)
            header_props = self.generate_props({AMQPProviderConfig.X_MATCH: "all", str(queue.header_value_match): True})
            self.consumer_channel.queue_bind(queue.return_name, AMQPProviderConfig.RETRY_EXCHANGE, AMQPProviderConfig.DEFAULT_ROUTING_KEY, header_props)

    def setup_queue(self, queue: QueueSubscribed) -> None:
        self.declare_queue(queue.name)
        if queue.bind:
            bind_options = queue.bind
            self.declare_exchange(bind_options.exchange)
            self.bind_queue(queue.name, bind_options.exchange.name, bind_options.routing_key or "")
            print(
                f"BindQueue - channelId={self.consumer_channel.channel_number} queue={queue.name} exchange={bind_options.exchange.name} routingKey={bind_options.routing_key}")
        else:
            print(f"No binding options - channelId={self.consumer_channel.channel_number} queue={queue.name}")
        self.consume_queue(queue)

    def declare_queue(self, queue_name: str) -> None:
        self.consumer_channel.queue_declare(queue_name, True, False, False, None)

    def declare_exchange(self, exchange: Exchange) -> None:
        self.consumer_channel.exchange_declare(exchange.name, exchange.exchange_type.value, durable=exchange.durable,
                                               auto_delete=exchange.auto_delete, arguments=self.generate_props({}))

    def init_connection(self, parameters: AMQPInit) -> pika.BlockingConnection:
        # connection_credentials = pika.PlainCredentials(
        #     username=parameters.username,
        #     password=parameters.password,
        #     erase_on_connect=True
        # )
        #
        # connection_parameters = pika.ConnectionParameters(
        #     host=parameters.hostname,
        #     port=parameters.port,
        #     virtual_host=parameters.virtual_host,
        #     credentials=connection_credentials
        # )
        connection = self.generate_connection_factory(parameters)
        print(f"New AMQP connection created. connectionId={connection}")
        return connection

    # def is_sensitive_content(self, delivery: pika.spec.Basic.Deliver) -> bool:
    #     return bool(delivery.properties.headers.get(AMQPProviderConfig.SENSITIVE_CONTENT_KEY, False))

    def consume_queue(self, queue: QueueSubscribed) -> None:
        print(
            f"Subscribing to queue={queue.name} with autoAckWhenCallback={queue.auto_ack_when_callback} channelId={self.consumer_channel.channel_number}")

        def cb(channel, method, properties, body):
            try:
                message = self.decode(body.decode("UTF-8"))

                content = message.content

                print(
                    f"Received Message - channelId={self.consumer_channel.channel_number} tag={method.delivery_tag} RoutingKey={method.routing_key} Exchange={method.exchange} commonId={message.common_id} content={content}")

                if queue.callback:
                    queue.callback(message, method)

                if queue.auto_ack_when_callback:
                    self.ack_message(method)
            except Exception as ex:
                print(
                    f"Error processing message. channelId={self.consumer_channel.channel_number} tag={method.delivery_tag} message={ex}")

        def cancel(consumer_tag):
            pass

        # Code below is useful to debug
        #
        # method_frame, header_frame, body = self.consumer_channel.basic_get(queue.name)
        # if method_frame:
        #     print("Received message from", queue.name)
        #     print("Message body:", body)
        #     self.consumer_channel.basic_ack(method_frame.delivery_tag)
        # else:
        #     print("No message returned from", queue.name)

        self.consumer_channel.basic_consume(queue.name, cb, False)
        print(f"Successfully enrolled in queue={queue.name} channelId={self.consumer_channel.channel_number}")

        try:
            self.consumer_channel.start_consuming()
        except KeyboardInterrupt:
            self.consumer_channel.stop_consuming()

    def bind_queue(self, queue_name: str, exchange_name: str, routing_key: str) -> None:
        self.consumer_channel.queue_bind(queue_name, exchange_name, routing_key)

    def publish(self, routing_key: str, message: str, common_id: str = str(uuid.uuid4()),
                encoded: bool = True, exchange: Optional[Exchange] = None, sensitive_content: bool = False) -> None:
        try:
            self.declare_destination(routing_key, exchange)

            # headers = self.generate_props({SensitiveContentKey: sensitive_content})

            headers = self.generate_props({})

            message_properties = BasicProperties(content_type=self.content_type_text, headers=headers, delivery_mode=2)

            self.publisher_channel.basic_publish(
                exchange=exchange.name if exchange else AMQPProviderConfig.DEFAULT_EXCHANGE,
                routing_key=routing_key,
                body=self.encode(message, common_id) if encoded else message.encode(),
                properties=message_properties
            )

            content = StringDecoder(message).decode if sensitive_content else message

            if exchange:
                print(
                    f"[Publish] message - channelId={self.publisher_channel.channel_number} exchange={exchange.name} type={exchange.exchange_type} routingKey={routing_key} commonId={common_id} content={content}")
            else:
                print(
                    f"[Publish] message - channelId={self.publisher_channel.channel_number} queue={routing_key} commonId={common_id} content={content}")

        except Exception as e:
            print(
                f"Failed to publish message to queue. channelId={self.publisher_channel.channel_number} commonId={common_id} error={e}")

    def declare_destination(self, queue_name: str, exchange: Optional[Exchange]) -> None:
        if self.destination_declared:
            return

        if exchange:
            self.declare_exchange(exchange)
        else:
            self.declare_queue(queue_name)

        self.destination_declared = True

    def ack_message(self, delivery: pika.spec.Basic.Deliver) -> None:
        tag = delivery.delivery_tag
        print(f"Ack message. channelId={self.consumer_channel.channel_number} tag={tag}")
        self.consumer_channel.basic_ack(delivery.delivery_tag, False)

    def send_back(self, delivery: pika.spec.Basic.Deliver, routing_key: Optional[str] = None) -> None:
        self.ack_message(delivery)

        attempts = delivery.get_properties().headers.get(AMQPProviderConfig.RETRY_COUNT_HEADER, 0)
        current_attempt = 1 if attempts >= len(self.RETRY_QUEUES) else attempts + 1

        print(
            f"[sendBack] channelId={self.consumer_channel.channel_number} MaxNoOfRetries={AMQPProviderConfig.MAX_NO_OF_RETRIES} attempts={attempts} currentAttempt={current_attempt} deliveryRoutingKey={delivery.envelope.routing_key} routingKey={routing_key} tag={delivery.envelope.delivery_tag}")

        if attempts >= AMQPProviderConfig.MAX_NO_OF_RETRIES:
            self.publish(
                f"{delivery.routing_key}.dead_letter",
                delivery.decode(delivery.get_body()),
                encoded=False
                # sensitive_content=self.is_sensitive_content(delivery)
            )
            print(
                f"Sent dead letter. channelId={self.publisher_channel.channel_number} queue={routing_key or delivery.routing_key} tag={delivery.delivery_tag}")
        else:
            headers = self.generate_props({
                str(current_attempt): True,
                AMQPProviderConfig.RETRY_COUNT_HEADER: (attempts + 1)
                # SensitiveContentKey: self.is_sensitive_content(delivery)
            })
            properties = BasicProperties(content_type="text/plain", headers=headers, delivery_mode=2)

            print(
                f"[sendBack] - Publishing to RetryExchange. channelId={self.publisher_channel.channel_number} attempts={attempts} currentAttempt={current_attempt} headers={headers} properties={properties}")

            self.publisher_channel.basic_publish(
                exchange=AMQPProviderConfig.RETRY_EXCHANGE,
                routing_key=routing_key or delivery.routing_key,
                body=bytes(delivery.get_body()),
                properties=properties
            )



from enum import Enum
from typing import Optional, Callable


class Published(Enum):
    OK_PUBLISHED = 1
    ERROR_PUBLISHED = 2


class SubscriptionStatus(Enum):
    OK_SUBSCRIPTION = 1
    ERROR_SUBSCRIPTION = 2


class ExchangeType(Enum):
    TOPIC = "topic"
    HEADERS = "headers"
    FANOUT = "fanout"
    DIRECT = "direct"

    def exchange_type_to_string(self) -> str:
        return self.value


class Exchange:
    def __init__(
            self,
            name: str,
            exchange_type: ExchangeType = ExchangeType.TOPIC,
            durable: bool = True,
            auto_delete: bool = False,
            internal: bool = False
    ) -> None:
        self.name: str = name
        self.exchange_type: ExchangeType = exchange_type
        self.durable: bool = durable
        self.auto_delete: bool = auto_delete
        self.internal: bool = internal


class BindOptions:
    def __init__(self, exchange: Exchange, routing_key: Optional[str] = None) -> None:
        self.exchange: Exchange = exchange
        self.routing_key: Optional[str] = routing_key


class QueueSubscribed:
    def __init__(
            self,
            name: str,
            bind: Optional[BindOptions] = None,
            auto_ack_when_callback: bool = False,
            callback: Optional[Callable] = None
    ) -> None:
        self.name: str = name
        self.bind: Optional[BindOptions] = bind
        self.auto_ack_when_callback: bool = auto_ack_when_callback
        self.callback: Optional[Callable] = callback



"""
tamqp.

An adaptable Python library to interact with your choice of cloud RabbitMQ providrs. Originally designed and developed at Taekus.
"""

__version__ = "0.0.1"
__author__ = 'Saurabh Gupta'
__credits__ = 'Taekus Inc.'


import pika
import ssl
from typing import Any, Dict
from tamqp.models.amqp_init import AMQPInit


class AMQPUtils:
    def generate_props(self, input: Dict[str, Any]) -> Dict[str, Any]:
        properties = {}
        for key, value in input.items():
            properties[key] = value
        return properties

    def generate_connection_factory(self, parameters):
        credentials = pika.PlainCredentials(parameters.username, parameters.password)

        ssl_options = None
        if parameters.is_tls:
            ssl_options = pika.SSLOptions(ssl.create_default_context(ssl.Purpose.CLIENT_AUTH))

        params = pika.ConnectionParameters(
            host=parameters.hostname,
            port=parameters.port,
            virtual_host=parameters.virtual_host or '/',
            credentials=credentials,
            ssl_options=ssl_options
        )

        return pika.BlockingConnection(params)



from setuptools import setup, find_packages

setup(
    name='tamqp',
    version='0.0.2',
    description='Adaptable Python library to interact with cloud RabbitMQ providers',
    url='https://s-g@bitbucket.org/yployalty/tamqp',
    author='Saurabh Gupta',
    author_email='sgupta@taekus.com',
    license='GNU General Public License v3.0',
    packages=find_packages(),
    install_requires=['pika==1.3.2', ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
    ],
)



class RetryQueue:
    def __init__(self, header_value_match: int, ttl: int, dlx_exchange: str = ""):
        self.header_value_match: int = header_value_match
        self.ttl: int = ttl
        self.dlx_exchange: str = dlx_exchange

    @property
    def return_name(self) -> str:
        return f"retry.queue.{self.header_value_match}"
